@@ -137,17 +137,32 @@ static bool exec_and_wait( const char *const argv[] )
     _exit(execv(argv[0], const_cast<char *const *>(argv)));
 }
 
+// Removes everything in the directory recursively, but not the directory itself.
+static bool cleanup_directory( const string& directory )
+{
+    DIR* dir = opendir( directory.c_str());
+    if( dir == NULL )
+        return false;
+    while( dirent* f = readdir( dir )) {
+        if( strcmp( f->d_name, "." ) == 0 || strcmp( f->d_name, ".." ) == 0 )
+            continue;
+        string fullpath = directory + '/' + f->d_name;
+        if( unlink( fullpath.c_str()) != 0 ) {
+            if( !cleanup_directory( fullpath ) || rmdir( fullpath.c_str()) != 0 ) {
+                return false;
+            }
+        }
+    }
+    closedir( dir );
+    return true;
+}
+
 bool cleanup_cache( const string &basedir )
 {
     flush_debug();
 
-    // make sure it ends with '/' to not fall into symlink traps
-    string bdir = basedir + '/';
-    const char *const argv[] = {
-        "/bin/rm", "-rf", "--", bdir.c_str(), NULL
-    };
-
-    bool ret = exec_and_wait( argv );
+    if ( access( basedir.c_str(), R_OK ) == 0 && !cleanup_directory( basedir ))
+        return false;
 
     if ( mkdir( basedir.c_str(), 0755 ) && errno != EEXIST ) {
         if ( errno == EPERM )
@@ -157,7 +172,7 @@ bool cleanup_cache( const string &basedir )
         return false;
     }
 
-    return ret;
+    return true;
 }
 
 Environments available_environmnents(const string &basedir)
@@ -247,7 +262,7 @@ size_t setup_env_cache(const string &basedir, string &native_environment, uid_t
     if ( mkdir( nativedir.c_str(), 0775 ) && errno != EEXIST )
    	return 0; 
 
-    if ( chown( nativedir.c_str(), 0, user_gid ) ||
+    if ( chown( nativedir.c_str(), user_uid, user_gid ) ||
          chmod( nativedir.c_str(), 0775 ) ) {
 	rmdir( nativedir.c_str() );
 	return 0;
@@ -362,7 +377,7 @@ pid_t start_install_environment( const std::string &basename, const std::string
         return 0;
     }
 
-    if ( chown( dirname.c_str(), 0, user_gid ) ||
+    if ( chown( dirname.c_str(), user_uid, user_gid ) ||
          chmod( dirname.c_str(), 0770 ) ) {
         log_perror( "chown,chmod target" );
         return 0;
@@ -374,7 +389,7 @@ pid_t start_install_environment( const std::string &basename, const std::string
         return 0;
     }
 
-    if ( chown( dirname.c_str(), 0, user_gid ) ||
+    if ( chown( dirname.c_str(), user_uid, user_gid ) ||
          chmod( dirname.c_str(), 0770 ) ) {
         log_perror( "chown,chmod name" );
         return 0;
@@ -433,7 +448,7 @@ pid_t start_install_environment( const std::string &basename, const std::string
 
 
 size_t finalize_install_environment( const std::string &basename, const std::string &target,
-                                     pid_t pid, gid_t user_gid)
+                                     pid_t pid, uid_t user_uid, gid_t user_gid)
 {
     int status = 1;
     while ( waitpid( pid, &status, 0) < 0 && errno == EINTR)
@@ -447,7 +462,7 @@ size_t finalize_install_environment( const std::string &basename, const std::str
 
     string dirname = basename + "/target=" + target;
     mkdir( ( dirname + "/tmp" ).c_str(), 01775 );
-    chown( ( dirname + "/tmp" ).c_str(), 0, user_gid );
+    chown( ( dirname + "/tmp" ).c_str(), user_uid, user_gid );
     chmod( ( dirname + "/tmp" ).c_str(), 01775 );
 
     return sumup_dir (dirname);
