@@ -1,4 +1,4 @@
-/*
+/* -*- c-file-style: "java"; indent-tabs-mode: nil -*-
     This file is part of Icecream.
 
     Copyright (c) 2004 Stephan Kulow <coolo@suse.de>
@@ -78,6 +78,10 @@
 #  define RUSAGE_CHILDREN (-1)
 #endif
 
+#ifdef HAVE_LIBCAP_NG
+#  include <cap-ng.h>
+#endif
+
 #include <deque>
 #include <map>
 #include <algorithm>
@@ -363,68 +367,42 @@ void usage(const char* reason = 0)
   if (reason)
      cerr << reason << endl;
 
-  cerr << "usage: iceccd [-n <netname>] [-m <max_processes>] [--no-remote] [-w] [-d|--daemonize] [-l logfile] [-s <schedulerhost>] [-v[v[v]]] [-r|--run-as-user] [-b <env-basedir>] [-u|--nobody-uid <nobody_uid>] [--cache-limit <MB>] [-N <node_name>]" << endl;
+  cerr << "usage: iceccd [-n <netname>] [-m <max_processes>] [--no-remote] [-w] [-d|--daemonize] [-l logfile] [-s <schedulerhost>] [-v[v[v]]] [-r|--run-as-user] [-u|--user-uid <user_uid>] [-b <env-basedir>] [--cache-limit <MB>] [-N <node_name>]" << endl;
   exit(1);
 }
 
-int setup_listen_fd()
-{
-    int listen_fd;
-    if ((listen_fd = socket (PF_INET, SOCK_STREAM, 0)) < 0) {
-        log_perror ("socket()");
-        return -1;
-    }
-
-    int optval = 1;
-    if (setsockopt (listen_fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) < 0) {
-        log_perror ("setsockopt()");
-        return -1;
-    }
-
-    int count = 5;
-    while ( count ) {
-        struct sockaddr_in myaddr;
-        myaddr.sin_family = AF_INET;
-        myaddr.sin_port = htons (PORT);
-        myaddr.sin_addr.s_addr = INADDR_ANY;
-        if (bind (listen_fd, (struct sockaddr *) &myaddr,
-                  sizeof (myaddr)) < 0) {
-            log_perror ("bind()");
-            sleep( 2 );
-            if ( !--count )
-                return -1;
-            continue;
-        } else
-            break;
-    }
-
-    if (listen (listen_fd, 20) < 0)
-    {
-      log_perror ("listen()");
-      return -1;
-    }
-
-    fcntl(listen_fd, F_SETFD, FD_CLOEXEC);
-
-    return listen_fd;
-}
-
-
 struct timeval last_stat;
 int mem_limit = 100;
 unsigned int max_kids = 0;
 
 size_t cache_size_limit = 100 * 1024 * 1024;
 
+struct NativeEnvironment
+{
+    string name; // the hash
+    map<string, time_t> extrafilestimes;
+    // Timestamps for compiler binaries, if they have changed since the time
+    // the native env was built, it needs to be rebuilt.
+    time_t gcc_bin_timestamp;
+    time_t gpp_bin_timestamp;
+    time_t clang_bin_timestamp;
+};
+
 struct Daemon
 {
     Clients clients;
     map<string, time_t> envs_last_use;
-    string native_environment;
+    // Map of native environments, the basic one(s) containing just the compiler
+    // and possibly more containing additional files (such as compiler plugins).
+    // The key is the compiler name and a concatenated list of the additional files
+    // (or just the compiler name for the basic ones).
+    map<string, NativeEnvironment> native_environments;
     string envbasedir;
-    uid_t nobody_uid;
-    gid_t nobody_gid;
-    int listen_fd;
+    uid_t user_uid;
+    gid_t user_gid;
+    bool warn_icecc_user;
+    int tcp_listen_fd;
+    int unix_listen_fd;
     string machine_name;
     string nodename;
     bool noremote;
@@ -449,10 +427,20 @@ struct Daemon
     unsigned int current_kids;
 
     Daemon() {
+        struct passwd *pw = getpwnam("icecc");
+        if (pw) {
+            user_uid = pw->pw_uid;
+            user_gid = pw->pw_gid;
+            warn_icecc_user = false;
+        } else {
+            warn_icecc_user = true;
+            user_uid = 65534;
+            user_gid = 65533;
+        }
+
         envbasedir = "/tmp/icecc-envs";
-        nobody_uid = 65534;
-        nobody_gid = 65533;
-        listen_fd = -1;
+        tcp_listen_fd = -1;
+	unix_listen_fd = -1;
         new_client_id = 0;
         next_scheduler_connect = 0;
         cache_size = 0;
@@ -474,7 +462,7 @@ struct Daemon
     int answer_client_requests();
     bool handle_transfer_env( Client *client, Msg *msg ) __attribute_warn_unused_result__;
     bool handle_transfer_env_done( Client *client );
-    bool handle_get_native_env( Client *client ) __attribute_warn_unused_result__;
+    bool handle_get_native_env( Client *client, GetNativeEnvMsg *msg ) __attribute_warn_unused_result__;
     void handle_old_request();
     bool handle_compile_file( Client *client, Msg *msg ) __attribute_warn_unused_result__;
     bool handle_activity( Client *client ) __attribute_warn_unused_result__;
@@ -487,6 +475,8 @@ struct Daemon
     bool handle_local_job( Client *client, Msg *msg ) __attribute_warn_unused_result__;
     bool handle_job_done( Client *cl, JobDoneMsg *m ) __attribute_warn_unused_result__;
     bool handle_compile_done (Client* client) __attribute_warn_unused_result__;
+    bool handle_verify_env ( Client* client, VerifyEnvMsg *msg ) __attribute_warn_unused_result__;
+    bool handle_blacklist_host_env ( Client* client, Msg *msg ) __attribute_warn_unused_result__;
     int handle_cs_conf( ConfCSMsg *msg);
     string dump_internals() const;
     string determine_nodename();
@@ -496,8 +486,87 @@ struct Daemon
     void close_scheduler();
     bool reconnect();
     int working_loop();
+    bool setup_listen_fds();
+    void check_cache_size( const string &new_env );
 };
 
+bool Daemon::setup_listen_fds()
+{
+    tcp_listen_fd = -1;
+    if (!noremote) { // if we only listen to local clients, there is no point in going TCP
+	if ((tcp_listen_fd = socket (PF_INET, SOCK_STREAM, 0)) < 0) {
+            log_perror ("socket()");
+            return false;
+        }
+
+        int optval = 1;
+        if (setsockopt (tcp_listen_fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) < 0) {
+            log_perror ("setsockopt()");
+            return false;
+        }
+        
+        int count = 5;
+        while ( count ) {
+            struct sockaddr_in myaddr;
+            myaddr.sin_family = AF_INET;
+            myaddr.sin_port = htons (PORT);
+            myaddr.sin_addr.s_addr = INADDR_ANY;
+            if (bind (tcp_listen_fd, (struct sockaddr *) &myaddr,
+                      sizeof (myaddr)) < 0) {
+                log_perror ("bind()");
+                sleep( 2 );
+                if ( !--count )
+                    return false;
+                continue;
+            } else
+                break;
+        }
+        
+        if (listen (tcp_listen_fd, 20) < 0) {
+            log_perror ("listen()");
+            return false;
+        }
+        
+        fcntl(tcp_listen_fd, F_SETFD, FD_CLOEXEC);
+    }
+    
+    if ((unix_listen_fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
+        log_perror ("socket()");
+        return false;
+    }
+    struct sockaddr_un myaddr;
+    memset(&myaddr, 0, sizeof(myaddr));
+    myaddr.sun_family = AF_UNIX;
+    mode_t old_umask = -1U;
+    if (access("/var/run/icecc", R_OK|W_OK|X_OK) == 0) {
+        strncpy(myaddr.sun_path, "/var/run/icecc/iceccd.socket", sizeof(myaddr.sun_path)-1);
+        unlink(myaddr.sun_path);
+        old_umask = umask(0);
+    } else {
+        strncpy(myaddr.sun_path, getenv("HOME"), sizeof(myaddr.sun_path)-1);
+        strncat(myaddr.sun_path, "/.iceccd.socket", sizeof(myaddr.sun_path)-1-strlen(myaddr.sun_path));
+        unlink(myaddr.sun_path);
+    }
+
+    if (bind(unix_listen_fd, (struct sockaddr*)&myaddr, sizeof(myaddr)) < 0) {
+        log_perror("bind()");
+        if (old_umask != -1U)
+            umask(old_umask);
+        return false;
+    }
+    if (old_umask != -1U)
+        umask(old_umask);
+
+    if (listen (unix_listen_fd, 20) < 0) {
+        log_perror ("listen()");
+        return false;
+    }
+    
+    fcntl(unix_listen_fd, F_SETFD, FD_CLOEXEC);
+    
+    return true;
+}
+
 void Daemon::determine_system()
 {
     struct utsname uname_buf;
@@ -654,8 +723,9 @@ string Daemon::dump_internals() const
     if ( cache_size )
         result += "  Cache Size: " + toString( cache_size ) + "\n";
     result += "  Architecture: " + machine_name + "\n";
-    if ( !native_environment.empty() )
-        result += "  NativeEnv: " + native_environment + "\n";
+    for (map<string, NativeEnvironment>::const_iterator it = native_environments.begin();
+         it!= native_environments.end(); ++it)
+        result += "  NativeEnv (" + it->first + "): " + it->second.name + "\n";
 
     if ( !envs_last_use.empty() )
         result += "  Now: " + toString( time( 0 ) ) + "\n";
@@ -736,8 +806,8 @@ bool Daemon::handle_transfer_env( Client *client, Msg *_msg )
     int sock_to_stdin = -1;
     FileChunkMsg* fmsg = 0;
 
-    pid_t pid = start_install_environment( envbasedir, emsg->target,
-        emsg->name, client->channel, sock_to_stdin, fmsg, nobody_uid, nobody_gid );
+    pid_t pid = start_install_environment( envbasedir, target,
+        emsg->name, client->channel, sock_to_stdin, fmsg, user_uid, user_gid );
 
     client->status = Client::TOINSTALL;
     client->outfile = emsg->target + "/" + emsg->name;
@@ -765,7 +835,7 @@ bool Daemon::handle_transfer_env_done( Client *client )
     assert(client->status == Client::TOINSTALL);
 
     size_t installed_size = finalize_install_environment(envbasedir, client->outfile,
-                              client->child_pid, nobody_gid);
+                              client->child_pid, user_uid, user_gid);
 
     if (client->pipe_to_child >= 0) {
         installed_size = 0;
@@ -788,17 +858,41 @@ bool Daemon::handle_transfer_env_done( Client *client )
         log_error() << "installed " << current << " size: " << installed_size
             << " all: " << cache_size << endl;
     }
+    check_cache_size( current );
 
+    bool r = reannounce_environments(); // do that before the file compiles
+    // we do that here so we're not given out in case of full discs
+    if ( !maybe_stats(true) )
+        r = false;
+    return r;
+}
+
+void Daemon::check_cache_size( const string &new_env )
+{
     time_t now = time( NULL );
     while ( cache_size > cache_size_limit ) {
         string oldest;
         // I don't dare to use (time_t)-1
         time_t oldest_time = time( NULL ) + 90000;
+        bool oldest_is_native = false;
         for ( map<string, time_t>::const_iterator it = envs_last_use.begin();
                 it != envs_last_use.end(); ++it ) {
             trace() << "das ist jetzt so: " << it->first << " " << it->second << " " << oldest_time << endl;
             // ignore recently used envs (they might be in use _right_ now)
-            if ( it->second < oldest_time && now - it->second > 200 ) {
+            int keep_timeout = 200;
+            bool native = false;
+            // If it is a native environment, allow removing it only after a longer period,
+            // unless there are many native environments.
+            for (map<string, NativeEnvironment>::const_iterator it2 = native_environments.begin();
+                 it2 != native_environments.end(); ++it2 ) {
+                if (it2->second.name == it->first) {
+                    native = true;
+                    if (native_environments.size() < 5)
+                        keep_timeout = 24 * 60 * 60; // 1 day
+                    break;
+                }
+            }
+            if ( it->second < oldest_time && now - it->second > keep_timeout ) {
                 bool env_currently_in_use = false;
                 for (Clients::const_iterator it2 = clients.begin(); it2 != clients.end(); ++it2)  {
                     if (it2->second->status == Client::TOCOMPILE ||
@@ -815,31 +909,63 @@ bool Daemon::handle_transfer_env_done( Client *client )
                 if (!env_currently_in_use) {
                     oldest_time = it->second;
                     oldest = it->first;
+                    oldest_is_native = native;
                 }
             }
         }
-        if ( oldest.empty() || oldest == current )
+        if ( oldest.empty() || oldest == new_env )
             break;
-        size_t removed = remove_environment( envbasedir, oldest );
-        trace() << "removing " << envbasedir << "/" << oldest << " " << oldest_time << " " << removed << endl;
+        size_t removed;
+        if (oldest_is_native) {
+            removed = remove_native_environment( oldest );
+            trace() << "removing " << oldest << " " << oldest_time << " " << removed << endl;
+        } else {
+            removed = remove_environment( envbasedir, oldest );
+            trace() << "removing " << envbasedir << "/" << oldest << " " << oldest_time << " " << removed << endl;
+        }
         cache_size -= min( removed, cache_size );
         envs_last_use.erase( oldest );
     }
-
-    bool r = reannounce_environments(); // do that before the file compiles
-    // we do that here so we're not given out in case of full discs
-    if ( !maybe_stats(true) )
-        r = false;
-    return r;
 }
 
-bool Daemon::handle_get_native_env( Client *client )
+bool Daemon::handle_get_native_env( Client *client, GetNativeEnvMsg *msg )
 {
-    trace() << "get_native_env " << native_environment << endl;
+    string env_key;
+    map<string, time_t> extrafilestimes;
+    env_key = msg->compiler;
+    for( list<string>::const_iterator it = msg->extrafiles.begin();
+         it != msg->extrafiles.end(); ++it ) {
+        env_key += ':';
+        env_key += *it;
+        struct stat st;
+        if( stat( it->c_str(), &st ) != 0 ) {
+            trace() << "Extra file " << *it << " for environment not found." << endl;
+            client->channel->send_msg( EndMsg() );
+            handle_end( client, 122 );
+            return false;
+        }
+        extrafilestimes[ *it ] = st.st_mtime;
+    }
 
-    if ( !native_environment.length() ) {
-        size_t installed_size = setup_env_cache( envbasedir, native_environment,
-                                                 nobody_uid, nobody_gid );
+    if ( native_environments[ env_key ].name.length()) {
+        const NativeEnvironment& env = native_environments[ env_key ];
+        if (!compilers_uptodate(env.gcc_bin_timestamp, env.gpp_bin_timestamp, env.clang_bin_timestamp)
+             || env.extrafilestimes != extrafilestimes
+             || access( env.name.c_str(), R_OK ) != 0) {
+            trace() << "native_env needs rebuild" << endl;
+            cache_size -= remove_native_environment( env.name );
+            envs_last_use.erase( env.name );
+            native_environments.erase( env_key ); // invalidates 'env'
+        }
+    }
+
+    trace() << "get_native_env " << native_environments[ env_key ].name
+        << " (" << env_key << ")" << endl;
+
+    if ( !native_environments[ env_key ].name.length()) {
+        NativeEnvironment& env = native_environments[ env_key ]; // also inserts it
+        size_t installed_size = setup_env_cache( envbasedir, env.name,
+                                                 user_uid, user_gid, msg->compiler, msg->extrafiles );
         // we only clean out cache on next target install
         cache_size += installed_size;
         trace() << "cache_size = " << cache_size << endl;
@@ -848,12 +974,17 @@ bool Daemon::handle_get_native_env( Client *client )
             handle_end( client, 121 );
             return false;
         }
+        env.extrafilestimes = extrafilestimes;
+        save_compiler_timestamps(env.gcc_bin_timestamp, env.gpp_bin_timestamp, env.clang_bin_timestamp);
+        envs_last_use[ native_environments[ env_key ].name ] = time( NULL );
+        check_cache_size( env.name );
     }
-    UseNativeEnvMsg m( native_environment );
+    UseNativeEnvMsg m( native_environments[ env_key ].name );
     if (!client->channel->send_msg( m )) {
         handle_end(client, 138);
         return false;
     }
+    envs_last_use[ native_environments[ env_key ].name ] = time( NULL );
     client->status = Client::GOTNATIVE;
     return true;
 }
@@ -926,7 +1057,7 @@ void Daemon::handle_old_request()
 
             string envforjob = job->targetPlatform() + "/" + job->environmentVersion();
             envs_last_use[envforjob] = time( NULL );
-            pid = handle_connection( envbasedir, job, client->channel, sock, mem_limit, nobody_uid, nobody_gid );
+            pid = handle_connection( envbasedir, job, client->channel, sock, mem_limit, user_uid, user_gid );
             trace() << "handle connection returned " << pid << endl;
 
             if ( pid > 0) {
@@ -1003,6 +1134,31 @@ bool Daemon::handle_compile_file( Client *client, Msg *msg )
     return true;
 }
 
+bool Daemon::handle_verify_env( Client *client, VerifyEnvMsg *msg )
+{
+    assert( msg );
+    bool ok = verify_env( client->channel, envbasedir, msg->target, msg->environment, user_uid, user_gid );
+    trace() << "Verify environment done, " << ( ok ? "success" : "failure" ) << ", environment " << msg->environment
+         << " (" << msg->target << ")" << endl;
+    VerifyEnvResultMsg resultmsg( ok );
+    if ( !client->channel->send_msg( resultmsg ))
+    {
+        log_error() << "sending verify end result failed.." << endl;
+        return false;
+    }
+    return true;
+}
+
+bool Daemon::handle_blacklist_host_env( Client *client, Msg *msg )
+{
+    // just forward
+    assert(dynamic_cast<BlacklistHostEnvMsg*>( msg ));
+    assert( client );
+    if ( !scheduler )
+        return false;
+    return send_scheduler( *msg );
+}
+
 void Daemon::handle_end( Client *client, int exitcode )
 {
 #ifdef ICECC_DEBUG
@@ -1209,13 +1365,15 @@ bool Daemon::handle_activity( Client *client )
     }
 
     switch ( msg->type ) {
-    case M_GET_NATIVE_ENV: ret = handle_get_native_env( client ); break;
+    case M_GET_NATIVE_ENV: ret = handle_get_native_env( client, dynamic_cast<GetNativeEnvMsg*>(msg) ); break;
     case M_COMPILE_FILE: ret = handle_compile_file( client, msg ); break;
     case M_TRANFER_ENV: ret = handle_transfer_env( client, msg ); break;
     case M_GET_CS: ret = handle_get_cs( client, msg ); break;
     case M_END: handle_end( client, 119 ); ret = false; break;
     case M_JOB_LOCAL_BEGIN: ret = handle_local_job (client, msg); break;
     case M_JOB_DONE: ret = handle_job_done( client, dynamic_cast<JobDoneMsg*>(msg) ); break;
+    case M_VERIFY_ENV: ret = handle_verify_env( client, dynamic_cast<VerifyEnvMsg*>(msg) ); break;
+    case M_BLACKLIST_HOST_ENV: ret = handle_blacklist_host_env( client, msg ); break;
     default:
         log_error() << "not compile: " << ( char )msg->type << "protocol error on client " << client->dump() << endl;
         client->channel->send_msg( EndMsg() );
@@ -1250,8 +1408,14 @@ int Daemon::answer_client_requests()
     struct timeval tv;
 
     FD_ZERO( &listen_set );
-    FD_SET( listen_fd, &listen_set );
-    int max_fd = listen_fd;
+    int max_fd = 0;
+    if (tcp_listen_fd != -1) {
+      FD_SET( tcp_listen_fd, &listen_set );
+      max_fd = tcp_listen_fd;
+    }
+    FD_SET( unix_listen_fd, &listen_set );
+    if (unix_listen_fd > max_fd) // very likely
+      max_fd = unix_listen_fd;
 
     for (map<int, MsgChannel *>::const_iterator it = fd2chan.begin();
          it != fd2chan.end();) {
@@ -1339,7 +1503,13 @@ int Daemon::answer_client_requests()
             }
         }
 
-        if ( FD_ISSET( listen_fd, &listen_set ) ) {
+	int listen_fd = -1;
+	if ( tcp_listen_fd != -1 && FD_ISSET( tcp_listen_fd, &listen_set ) )
+	  listen_fd = tcp_listen_fd;
+	if ( FD_ISSET( unix_listen_fd, &listen_set ) )
+	  listen_fd = unix_listen_fd;
+
+        if ( listen_fd != -1) {
             struct sockaddr cli_addr;
             socklen_t cli_len = sizeof cli_addr;
             int acc_fd = accept(listen_fd, &cli_addr, &cli_len);
@@ -1418,7 +1588,9 @@ bool Daemon::reconnect()
         return false;
     }
 
+#ifdef ICECC_DEBUG
     trace() << "reconn " << dump_internals() << endl;
+#endif
     if (!discover
 	|| discover->timed_out())
     {
@@ -1440,11 +1612,11 @@ bool Daemon::reconnect()
         remote_name = inet_ntoa( name.sin_addr );
     else
         remote_name = string();
-    log_info() << "Connected to scheduler (I am known as" << remote_name << ")\n";
+    log_info() << "Connected to scheduler (I am known as " << remote_name << ")\n";
     current_load = -1000;
     gettimeofday( &last_stat, 0 );
     icecream_load = 0;
-
+    
     LoginMsg lmsg( PORT, determine_nodename(), machine_name );
     lmsg.envs = available_environmnents(envbasedir);
     lmsg.max_kids = max_kids;
@@ -1491,7 +1663,7 @@ int main( int argc, char ** argv )
             { "name", 1, NULL, 'n'},
             { "scheduler-host", 1, NULL, 's' },
             { "env-basedir", 1, NULL, 'b' },
-            { "nobody-uid", 1, NULL, 'u'},
+            { "user-uid", 1, NULL, 'u'},
             { "cache-limit", 1, NULL, 0},
             { "no-remote", 0, NULL, 0},
             { 0, 0, 0, 0 }
@@ -1585,9 +1757,10 @@ int main( int argc, char ** argv )
                 if ( !pw ) {
                     usage( "Error: -u requires a valid username" );
                 } else {
-                    d.nobody_uid = pw->pw_uid;
-                    d.nobody_gid = pw->pw_gid;
-                    if (!d.nobody_gid || !d.nobody_uid) {
+                    d.user_uid = pw->pw_uid;
+                    d.user_gid = pw->pw_gid;
+                    d.warn_icecc_user = false;
+                    if (!d.user_gid || !d.user_uid) {
                         usage( "Error: -u <username> must not be root");
                     }
                 }
@@ -1600,16 +1773,41 @@ int main( int argc, char ** argv )
         }
     }
 
+    if (d.warn_icecc_user)
+        log_perror ("Error: no icecc user on system. Falling back to nobody.");
+
     umask(022);
 
-    if ( !logfile.length() && detach)
-        logfile = "/var/log/iceccd";
+    if (getuid() == 0) {
+        if (!logfile.length() && detach) {
+            if (mkdir("/var/log/icecc", S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)) {
+                if (errno == EEXIST) {
+                    chmod("/var/log/icecc", S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+                    chown("/var/log/icecc", d.user_uid, d.user_gid);
+                }
+            }
+
+            logfile = "/var/log/icecc/iceccd.log";
+        }
+
+        if (mkdir("/var/run/icecc", S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) == -1) {
+            if (errno == EEXIST) {
+                chmod("/var/run/icecc", S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+                chown("/var/run/icecc", d.user_uid, d.user_gid);
+            }
+        }
+#ifdef HAVE_LIBCAP_NG
+        capng_clear(CAPNG_SELECT_BOTH);
+        capng_update(CAPNG_ADD, (capng_type_t)(CAPNG_EFFECTIVE|CAPNG_PERMITTED), CAP_SYS_CHROOT);
+        capng_change_id(d.user_uid, d.user_gid, CAPNG_NO_FLAG);
+        capng_apply(CAPNG_SELECT_BOTH);
+#endif
+    } else {
+        d.noremote = true;
+    }
 
     setup_debug( debug_level, logfile );
 
-    if ((getuid()!=0))
-        d.noremote = true;
-
     log_info() << "ICECREAM daemon " VERSION " starting up (nice level "
                << nice_level << ") " << endl;
 
@@ -1673,9 +1871,9 @@ int main( int argc, char ** argv )
     for (list<string>::const_iterator it = nl.begin(); it != nl.end(); ++it)
         trace() << *it << endl;
 
-    d.listen_fd = setup_listen_fd();
-    if ( d.listen_fd == -1 ) // error
+    if ( !d.setup_listen_fds() ) // error
         return 1;
 
     return d.working_loop();
 }
+
